//ui.java Summer -Winter 2003
//Features:
// Show chat (and log)
// Show player list (with color)
// Show arena list*
// Send commands and chat to core for processing

import java.io.*;
import java.text.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import java.awt.*;
import java.awt.event.*;

import java.util.Hashtable;

public class ui
{
	static final int version = 9;

	static loggers lm;
	core    ml;
	data    pd;

	private JFrame frame;
	private JTextArea sheepText;
	//private JTextArea out;
	private JTextPane out;
	private AbstractDocument outDoc;
	private JList list;
	private DefaultListModel listModel;
	private JTextArea in;
	private JLabel statusLabel;
	private JLabel timeLabel;
	
	//Cell renderers
	//private ListRenderer myListRenderer;
	private Color renderColor;
	Hashtable plist;
	
	protected Color[] colors;
	final int GREY   = 0;
	final int GREEN  = 1;
	final int BLUE   = 2;
	final int RED    = 3;
	final int YELLOW = 4;
	final int PURPLE = 5;
	final int ORANGE = 6;
	final int PINK   = 7;
	
	//Quick ::
	static final int MAX_LAST = 5;
	static String[] lastName;
	static int lastIndex = 0;
	
	public ui(loggers lm, data pd)
	{
		this.lm = lm;
		this.ml = null;
		this.pd = pd;
		plist = new Hashtable();
		
		//look and feel
		try 
		{
			UIManager.setLookAndFeel(
				UIManager.getCrossPlatformLookAndFeelClassName()
				);
		}
		catch (Exception e)
		{
		}

		colors = new Color[8];
		colors[0] = new Color(189, 181, 181); //grey
		colors[1] = new Color(107, 222,  90); //green
		colors[2] = new Color(181, 181, 255); //blue
		colors[3] = new Color(222,  49,   8); //red
		colors[4] = new Color(255, 189,  41); //yellow
		colors[5] = new Color(181, 132, 239); //purple
		colors[6] = new Color(239, 107,  24); //orange
		colors[7] = new Color(255, 140, 156); //pink

		lastName = new String[MAX_LAST];
		for(int i=0; i<MAX_LAST; i++)
			lastName[i] = new String("");

		//If I do this the linux wm loses control
		//JFrame.setDefaultLookAndFeelDecorated(true);

		//initialise window
		frame = new JFrame("chatnet");
		Component contents = createComponents();
		frame.getContentPane().add(contents, BorderLayout.CENTER);

		//Finish setting up the frame, and show it.
		frame.addWindowListener(
			new WindowAdapter()
			{
 				public void windowClosing(WindowEvent e)
				{
					//System.exit(0);
					ml.stage = ml.STAGE_QUIT;
				}
			}
			);
		//Use code above to save log
		//frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		
		//Center on screen
/*
		Dimension size = frame.getMinimumSize();
		lm.Log("D", "w: " + size.width + ", h: " + size.height);
		Dimension res = Toolkit.getDefaultToolkit().getScreenSize();
		lm.Log("D", "w: " + res.width + ", h: " + res.height);
		frame.setLocation(
			(res.width - size.width)/2,
			(res.height - size.height)/2);
*/
		frame.setLocation(30, 30);
		frame.pack();
		//frame.setVisible(true);
	}

	public void setCore(core ml)
	{
		this.ml = ml;
	}

	public Component createComponents()
	{
		//Create the components to go into the frame...
/*
----------------------
|Sheep message       |
----------------------
|Chat area   |player |
|            |list   |
|            |       |
|            |       |
|--------------------
|Input area          |
|                    |
----------------------
|Status bar          |
----------------------
*/
		//Sheep message
		JLabel sheepLabel = new JLabel("Sheep: ");
		//sheepLabel.setBorder(
		//	BorderFactory.createEmptyBorder(0, 3, 0, 3)
		//	);
		
		sheepText = new JTextArea();
		//sheepText.setBorder(
		//	BorderFactory.createEmptyBorder(0, 2, 0, 2)
		//	);
		sheepText.setRows(1);
		sheepText.setFont(new Font("Monospaced", Font.PLAIN, 11));
		sheepText.setBackground(Color.BLACK);
		sheepText.setForeground(colors[this.GREEN]);
		sheepText.setEditable(false);
		
		//Output text area
		//out = new JTextArea("Type ?exit to quit", 14, 52);
		//out.setFont(new Font("Monospaced", Font.PLAIN, 11));
		//out.setEditable(false);
		//out.setLineWrap(true);

		out = new JTextPane();
		out.setEditable(false);
		out.setBackground(Color.BLACK);
		out.setCaretPosition(0); //eh?
		//out.setMargin(new Insets(5,5,5,5));
		StyledDocument styledDoc = out.getStyledDocument();
		if (styledDoc instanceof AbstractDocument)
		{
			outDoc = (AbstractDocument)styledDoc;
			try
			{
				SimpleAttributeSet attr = new SimpleAttributeSet();
				StyleConstants.setFontFamily(attr, "Monospaced");
				StyleConstants.setFontSize(attr, 11);
				StyleConstants.setForeground(attr, colors[this.GREEN]);
				outDoc.insertString(outDoc.getLength(), "Type ?exit to quit", attr);
			}
			catch (BadLocationException ble)
			{
				lm.Log("E", "Couldn't insert initial text.");
				lm.Log("E", ble.getMessage());
		}
		} else
		{
			lm.Log("E", "Text pane's document isn't an AbstractDocument!");
			ml.disconnect();
		}

		//Player list
		listModel = new DefaultListModel();
		list = new JList(listModel);
		list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		list.setLayoutOrientation(JList.VERTICAL);
		list.setVisibleRowCount(8);
		list.setFixedCellWidth(110);
		list.setFont(new Font("Monospaced", Font.PLAIN, 11));
		list.setBackground(Color.BLACK);
		list.setForeground(colors[this.YELLOW]);
		list.setCellRenderer(new ListRenderer());
		//list.setSelectionBackground(Color.BLACK);
		//list.setSelectionForeground(colors[this.RED]);
		
		//Input text area
		in = new JTextArea();
		in.setRows(2);
		in.setFont(new Font("Monospaced", Font.PLAIN, 11));
		in.setBackground(Color.BLACK);
		in.setForeground(colors[this.BLUE]);
		in.setCaretColor(colors[this.BLUE]);
		in.setLineWrap(true);
		in.getDocument().addDocumentListener(
			new uiChatInputDocListener()
			);

		//Status bar label
		statusLabel = new JLabel("Ready");
		statusLabel.setBorder(
			BorderFactory.createEmptyBorder(0, 3, 0, 3)
			);

		timeLabel = new JLabel("[00:00]");

		JPanel pane;
		JPanel upperStatusPane;
		JPanel chatAndPlayerPane;
		JPanel chatInputPane;
		JPanel lowerStatusPane;
		JScrollPane scroller;
		
		//...stick them in a container
		pane = new JPanel();
		pane.setLayout(new BoxLayout(pane, BoxLayout.Y_AXIS));
		pane.setBorder(
			BorderFactory.createEmptyBorder(
				2, //Top
				2, //Left
				2, //Bottom
				2) //Right
			);
		//pane.setLayout(new BoxLayout(pane, BoxLayout.Y_AXIS));
		
		upperStatusPane   = new JPanel();
		upperStatusPane.setLayout(new BoxLayout(upperStatusPane, BoxLayout.X_AXIS));
		chatAndPlayerPane = new JPanel();
		chatAndPlayerPane.setLayout(new BoxLayout(chatAndPlayerPane, BoxLayout.X_AXIS));
		chatInputPane     = new JPanel(new GridLayout(1, 1));
		lowerStatusPane   = new JPanel();
		lowerStatusPane.setLayout(new BoxLayout(lowerStatusPane, BoxLayout.X_AXIS));
		
		//Add sub-panes
		pane.add(upperStatusPane);
		pane.add(chatAndPlayerPane);
		pane.add(chatInputPane);
		pane.add(lowerStatusPane);
		
		//Add controls
		//upperStatusPane
		upperStatusPane.setBorder(
			BorderFactory.createEmptyBorder(0, 0, 2, 0)
			);
		upperStatusPane.setMaximumSize(
			new Dimension(Integer.MAX_VALUE, 20)
			);
		upperStatusPane.add(sheepLabel);
		upperStatusPane.add(sheepText);
		
		//chatAndPlayerPane
		scroller = new JScrollPane(out,
			JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
			JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		scroller.setPreferredSize(new Dimension(384, 200));
		chatAndPlayerPane.add(scroller);

		scroller = new JScrollPane(list,
			JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
			JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		scroller.setBorder(
			BorderFactory.createEmptyBorder(0, 2, 1, 0)
			);
		scroller.setMaximumSize(
			new Dimension(110, Integer.MAX_VALUE)
			);
		chatAndPlayerPane.add(scroller);

		//chatInputPane
		chatInputPane.setMaximumSize(
			new Dimension(Integer.MAX_VALUE, 42)
			);
		chatInputPane.setBorder(
			BorderFactory.createEmptyBorder(1, 0, 0, 0)
			);
		scroller = new JScrollPane(in,
			JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
			JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		chatInputPane.add(scroller);
		
		//lowerStatusPane
		lowerStatusPane.add(statusLabel);
		lowerStatusPane.add(Box.createHorizontalGlue());
		lowerStatusPane.add(timeLabel);
		
		return pane;
	}

	class uiChatInputDocListener implements DocumentListener
	{
		public void insertUpdate(DocumentEvent e)
		{
			Runnable updateAComponent = new uiChatInputThreadSafe();
			SwingUtilities.invokeLater(updateAComponent);
		}
		
		public void removeUpdate(DocumentEvent e) {}
		public void changedUpdate(DocumentEvent e) {}
	}

	class uiChatInputThreadSafe implements Runnable
	{
		public void run()
		{
			String str = in.getText();
			
			//Chat colours
			if (str.startsWith("//") || str.startsWith("'"))
			{
				in.setForeground(colors[YELLOW]);
				in.setCaretColor(colors[YELLOW]);
			}
			else if (str.startsWith("/"))
			{
				in.setForeground(colors[GREEN]);
				in.setCaretColor(colors[GREEN]);
			}
			else if (str.startsWith(";"))
			{
				in.setForeground(colors[ORANGE]);
				in.setCaretColor(colors[ORANGE]);
			}
			else if (str.startsWith("\\"))
			{
				in.setForeground(colors[RED]);
				in.setCaretColor(colors[RED]);
			}
			else
			{
				in.setForeground(colors[BLUE]);
				in.setCaretColor(colors[BLUE]);
			}
			
			//Reset ::
			if (str.length() < 2)
				lastIndex = 0;
			
			//if (str.endsWith("\n") || str.endsWith("\r"))
			if(str.indexOf("\n") >= 0 || str.indexOf("\r") >= 0)
			{
				//Strip off all CRLF
				//str = str.substring(0, str.length()-1);
				str = str.replaceAll("\n", "");
				str = str.replaceAll("\r", "");
				
				if (!str.trim().equals(""))
				{
					if (str.startsWith("?") ||
						str.startsWith("*"))
					{
						ml.cmdHandler(str);
					}
					else
					{
						ml.chatHandler(str);
					}
				}

				in.setText("");
			}
			else if (str.endsWith("::"))
			{
				//hack
				String[] part = str.split(":", 4);
				if (part.length > 2 && part[2].length() == 0)
				
				if (str.startsWith(":"))
				{
					String name = getLastPrivPlayer();
					if (name.length() > 0)
						in.setText(":" + name + ":");
				}
			}
		}
	}

	public void show()
	{
		frame.setVisible(true);
		in.requestFocus();
	}

	public void hide()
	{
		frame.setVisible(false);
	}
	
	public void setSheep(String sheep)
	{
		sheepText.setText(sheep);
	}

	public String getSelectedName()
	{
		return (String)list.getSelectedValue();
	}

	public void addLastPrivPlayer(String name)
	{
		int i;

		//Already the first item? Return
		if (lastName[0].equalsIgnoreCase(name))
			return;

		//Shift down
		for (i=MAX_LAST-1; i>0; i--)
		{
			lastName[i] = lastName[i-1];
		}

		//Insert at top
		lastName[0] = name;

		//Already in there? Remove
		for (i=1; i<MAX_LAST; i++)
		{
			if (lastName[i].equalsIgnoreCase(name))
				lastName[i] = "";
		}
		
		lastIndex = 0;
	}



	public String getLastPrivPlayer()
	{
		return getLastPrivPlayer(lastIndex);
	}

	public String getLastPrivPlayer(int index)
	{
		int i;
		String name;
		
		if (index != lastIndex)
			lastIndex = index;
		
		for(i=0; i<MAX_LAST; i++)
		{
			name = lastName[lastIndex];
			lastIndex = (lastIndex + 1) % MAX_LAST;
			if (name != null && name.length() > 0) return name;
		}

		return "";
	}

	public void println(int color1, String name, int color2, String msg)
	{
		if (color1 < 0) color1 = 0;
		if (color1 > 7) color1 = 7;
		if (color2 < 0) color2 = 0;
		if (color2 > 7) color2 = 7;
		
		try
		{
			SimpleAttributeSet attr = new SimpleAttributeSet();
			StyleConstants.setFontFamily(attr, "Monospaced");
			StyleConstants.setFontSize(attr, 11);
			StyleConstants.setForeground(attr, colors[color1]);
			outDoc.insertString(outDoc.getLength(), "\n" + name, attr);
			StyleConstants.setForeground(attr, colors[color2]);
			outDoc.insertString(outDoc.getLength(), msg, attr);
			out.setCaretPosition(outDoc.getLength());
		}
		catch (BadLocationException ble)
		{
			lm.Log("E", "Couldn't insert text.");
			lm.Log("E", ble.getMessage());
		}

		System.out.println(name + msg);
		lm.Log(name + msg);
		//updateTime();
	}

	public void println(int color, String text)
	{
		if (color < 0) color = 0;
		if (color > 7) color = 7;
		
		try
		{
			SimpleAttributeSet attr = new SimpleAttributeSet();
			StyleConstants.setFontFamily(attr, "Monospaced");
			StyleConstants.setFontSize(attr, 11);
			StyleConstants.setForeground(attr, colors[color]);
			outDoc.insertString(outDoc.getLength(), "\n" + text, attr);
			out.setCaretPosition(outDoc.getLength());
		}
		catch (BadLocationException ble)
		{
			lm.Log("E", "Couldn't insert text.");
			lm.Log("E", ble.getMessage());
		}

		System.out.println(text);
		lm.Log(text);
		//updateTime();
	}

	public void println(String text)
	{
		println(this.GREEN, text);
	}
	
	public void printStatus(String text)
	{
		statusLabel.setText(text);
		updateTime();
	}

	public void clear()
	{
		out.setText("");
		try
		{
			SimpleAttributeSet attr = new SimpleAttributeSet();
			StyleConstants.setFontFamily(attr, "Monospaced");
			StyleConstants.setFontSize(attr, 11);
			StyleConstants.setForeground(attr, colors[ this.GREEN ]);
			outDoc.insertString(outDoc.getLength(), "Chat buffer cleared", attr);
			out.setCaretPosition(outDoc.getLength());
		}
		catch (BadLocationException ble)
		{
			lm.Log("E", "Couldn't insert text.");
			lm.Log("E", ble.getMessage());
		}
	}

	public void updateTime()
	{
		SimpleDateFormat formatter;
		formatter = new SimpleDateFormat("H:mm", new java.util.Locale("en","US").getDefault());
		timeLabel.setText("[" + formatter.format(new java.util.Date()) + "]");
		//System.out.println(new java.util.Locale("de","DE").getDefault().toString());
	}
	//Player list management
	public void playerEntered(String name, String freq)
	{
/*
		if (freq.equals(ml.freq))
		{
			//Render yellow
			renderColor = colors[this.YELLOW];
		}
		else
		{
			//Render grey
			renderColor = colors[this.GREY];
		}
*/	
		listModel.addElement(name);
		plist.put(name, freq);
	}
	
	public void playerLeft(String name)
	{
		listModel.removeElement(name);
		plist.remove(name);
	}
	
	public void playerReset()
	{
		listModel.clear();
		plist = new Hashtable();
		//playerEntered("1234512345 0000");
	}
	
	public void playerFreqChange(String name, String freq)
	{
		//Remove
		playerLeft(name);
		
		//Add
		playerEntered(name, freq);
	}
	
	class ListRenderer extends JLabel implements ListCellRenderer
	{
		Color yellow, grey;
		
		public ListRenderer()
		{
			super();
			
			yellow = new Color(255, 189,  41); //yellow
			grey   = new Color(189, 181, 181); //grey
		}
	
		public Component getListCellRendererComponent(
			JList list,
			Object value,		// value to display
			int index,		// cell index
			boolean isSelected,	// is the cell selected
			boolean cellHasFocus)	// the list and the cell have the focus
		{
			String s = value.toString();
 
 			if (isSelected)
			{
				setText("|" + s);
				//setBackground(list.getSelectionBackground());
				//setForeground(list.getSelectionForeground());
			}
			else
			{
				setText(s);
			}
			setBackground(list.getBackground());
			String f = (String)plist.get(s);
				
			if (f != null && f.equals(ml.freq))
				setForeground(yellow);
			else
				setForeground(grey);
			
			setEnabled(list.isEnabled());
			setFont(list.getFont());
			setOpaque(true);
			return this;
		}
	}
}
